#include <Arduino.h>
#include <WiFi.h>
#include <LITTLEFS.h>

#include "main.h"
#include "config.h"

const char *seqsTxt[] = {"CROSS", "SQUARE", "TRIANGLE", "CIRCLE", "L1", "R1", "L2", "R2", "L3", "R3", "START", "SELECT", NULL};

Config::Config()
{
	setDefaultValues();
}

void Config::setup()
{
	int saveDefaults=0;
	pinMode(RESETPIN, INPUT_PULLUP); // reset conf as input
	setDefaultValues();

	if (LITTLEFS.begin(true))
	{
		Serial.println(F("LITTLEFS mounted correctly."));
	}
	else
	{
		Serial.println(F("Cannot mount LITTLEFS !"));
	}

	if (!testResetConf())
	{
		if(!loadConfig())
		{
			// format because it can be a problem of flash size modification
			LITTLEFS.format();
			saveDefaults=1;
		}
		if (isTmpAp())
		{
			tmpAp = true;
		}
	}
	else
	{
		saveDefaults=1;
	}
	if(saveDefaults)
	{
		Serial.println(F("Setting default values..."));
		saveConfig();
	}
}

bool Config::testResetConf()
{
	return (digitalRead(RESETPIN) == 0);
}

String Config::getDefaultSsid()
{
	byte mac[6];
	WiFi.macAddress(mac);
	return "zenAmiSix-" + String(mac[4], HEX) + String(mac[5], HEX);
}

void Config::setDefaultValues()
{
	// default values
	ps3Mac = PS3MACADDR;
	ssid = getDefaultSsid();
	password = DEFPASSWD;
	numPlayer = JOY_DEF_NUMPLAYER;
	netName = DEFNETNAME;
	inactivityTimeoutS = DEFAULT_JOY_INACTIVITY_TIMEOUT_S;
	isAp = true;
	mouseCapable = false;
	tmpAp = false;
	for (int i = 0; i < SEQ_LAST; i++)
	{
		sequences_on[i] = sequences_off[i] = "";
	}
	sequences_on[SEQ_cross] = "F1";
	sequences_off[SEQ_cross] = "f1";

	sequences_on[SEQ_square] = "F2";
	sequences_off[SEQ_square] = "f2";

	sequences_on[SEQ_circle] = "F1T100f1T100+";
	sequences_off[SEQ_circle] = "f1";

	sequences_on[SEQ_r1] = "F1T10f1T10+";
	sequences_off[SEQ_r1] = "f1";
}

bool Config::loadConfig()
{
	// default values :
	setDefaultValues();

	File file = LITTLEFS.open("/config");
	if (!file || file.isDirectory())
	{
		file.close();
		return false;
	}
	String content;
	while (file.available())
	{
		int res=file.read();
		if(res==-1)
		{
			Serial.println(F("config file read error"));
			file.close();
			return false;
		}
		content += char(res);
	}
	file.close();

	// Serial.println("File content:"+content);
	JsonDocument doc;
	DeserializationError error = deserializeJson(doc, content);
	if (error)
	{
		Serial.print(F("deserializeJson() failed: "));
		Serial.println(error.c_str());
		return false;
	}
	parseFromJson(doc);
	Serial.println("Conf file loaded.");
	return true;
}

bool Config::saveConfig()
{
	Serial.println("Saving to file...");
	File file = LITTLEFS.open("/config", "w");
	if (!file)
	{
		Serial.println("Failed to save file");
		return false;
	}

	String toSave = this->toString();
	file.println(toSave);
	file.close();

	Serial.println("Conf file saved.");

	return true;
}

bool Config::parseFromJson(JsonDocument &doc)
{
	bool needToSaveToFile = false;

	if (doc.containsKey("MAC"))
	{
		const char *macFromDoc = doc["MAC"];
		String newMac = macFromDoc;
		if (ps3Mac != newMac)
		{
			ps3Mac = newMac;
			Serial.println("Asked to change MAC ADDR to " + ps3Mac);
			needToSaveToFile = true;
		}
	}
	if (doc.containsKey("SSID"))
	{
		const char *ssidFromDoc = doc["SSID"];
		String newSsid = ssidFromDoc;
		if (newSsid != ssid)
		{
			ssid = newSsid;
			Serial.println("Asked to change SSID to " + ssid);
			needToSaveToFile = true;
		}
	}
	if (doc.containsKey("PASS"))
	{
		const char *passFromDoc = doc["PASS"];
		String newPass = passFromDoc;
		if (newPass != password)
		{
			password = newPass;
			Serial.println("Asked to change Password to " + password);
			needToSaveToFile = true;
		}
	}
	if (doc.containsKey("NETNAME"))
	{
		const char *netNameFromDoc = doc["NETNAME"];
		String newNetName = netNameFromDoc;
		if (newNetName != netName)
		{
			netName = newNetName;
			Serial.print("Asked to change netName to " + netName);
			needToSaveToFile = true;
		}
	}
	if (doc.containsKey("ISAP"))
	{
		bool newIsAp = doc["ISAP"];
		if (newIsAp != isAp)
		{
			isAp = newIsAp;
			Serial.print("Asked to change isAp to ");
			Serial.println(isAp);
			needToSaveToFile = true;
		}
	}
	if (doc.containsKey("NUMPLAYER"))
	{
		int newNumPlayer = doc["NUMPLAYER"];
		if (newNumPlayer != numPlayer)
		{
			numPlayer = newNumPlayer;
			Serial.print("Asked to change numPlayer to ");
			Serial.println(numPlayer);
			needToSaveToFile = true;
		}
	}
	if (doc.containsKey("JOYTIMEOUT"))
	{
		int newInactivityTimeoutS = doc["JOYTIMEOUT"];
		if (newInactivityTimeoutS != inactivityTimeoutS)
		{
			inactivityTimeoutS = newInactivityTimeoutS;
			Serial.print("Asked to change inactivityTimeoutS to ");
			Serial.println(inactivityTimeoutS);
			needToSaveToFile = true;
		}
	}
	if (doc.containsKey("MOUSECAP"))
	{
		bool newMouseCap = doc["MOUSECAP"];
		if (newMouseCap != mouseCapable)
		{
			mouseCapable = newMouseCap;
			Serial.print("Asked to change mouseCapable to ");
			Serial.println(mouseCapable);
			needToSaveToFile = true;
		}
	}

	// sequences :
	int i;

	for (i = 0; seqsTxt[i]; i++)
	{
		String nameOn = "SEQ_" + String(seqsTxt[i]) + "_ON";
		String nameOff = "SEQ_" + String(seqsTxt[i]) + "_OFF";
		if (doc.containsKey(nameOn))
		{
			const char *newSeqCh = doc[nameOn];
			String newSeqStr = newSeqCh;
			if (sequences_on[i] != newSeqStr)
			{
				sequences_on[i] = newSeqStr;
				needToSaveToFile = true;
				Serial.print("Asked to change " + nameOn + " to ");
				Serial.println(newSeqStr);
			}
		}
		if (doc.containsKey(nameOff))
		{
			const char *newSeqCh = doc[nameOff];
			String newSeqStr = newSeqCh;
			if (sequences_off[i] != newSeqStr)
			{
				sequences_off[i] = newSeqStr;
				needToSaveToFile = true;
				Serial.print("Asked to change " + nameOff + " to ");
				Serial.println(newSeqStr);
			}
		}
	}

	if (doc.containsKey("REBOOT"))
	{
		Serial.println("Asked to REBOOT");
		needReboot = true;
	}
	if (doc.containsKey("RESETCONF"))
	{
		Serial.println("Asked to RESETCONF");
		setDefaultValues();
		saveConfig();
		needToSaveToFile = false;
		needReboot = true;
	}
	if (needToSaveToFile)
	{
		configChanged = true;
	}
	return needToSaveToFile;
}

// serialize
String Config::toString()
{
	String ret = "{";
	ret += "\"COMPILEDATE\":\"" + String(__DATE__) + "\"";
	ret += ",\"VERSION\":\"" + String(VERSION) + "\"";
	ret += ",\"MAC\":\"" + ps3Mac + "\"";
	ret += ",\"SSID\":\"" + ssid + "\"";
	ret += ",\"PASS\":\"" + password + "\"";
	ret += ",\"NETNAME\":\"" + netName + "\"";
	ret += ",\"NUMPLAYER\":" + String(numPlayer);
	ret += ",\"JOYTIMEOUT\":" + String(inactivityTimeoutS);
	if (isAp)
		ret += ",\"ISAP\":true";
	else
		ret += ",\"ISAP\":false";
	if (mouseCapable)
		ret += ",\"MOUSECAP\":true";
	else
		ret += ",\"MOUSECAP\":false";

	// sequences :

	int i;
	for (i = 0; seqsTxt[i]; i++)
	{
		String nameOn = "SEQ_" + String(seqsTxt[i]) + "_ON";
		String nameOff = "SEQ_" + String(seqsTxt[i]) + "_OFF";

		ret += ",\"" + nameOn + "\":\"" + sequences_on[i] + "\"";
		ret += ",\"" + nameOff + "\":\"" + sequences_off[i] + "\"";
	}

	if (amiga5VPresence)
		ret += ",\"AMIGAPRES\":true";
	else
		ret += ",\"AMIGAPRES\":false";

	ret += "}";

	// Serial.print("SER:");
	// Serial.println(ret);
	return ret;
}

bool Config::isConfigChanged()
{
	bool ret = configChanged;
	configChanged = false;
	return ret;
}

bool Config::isNeedReboot()
{
	return needReboot;
}

void Config::setNeedReboot()
{
	needReboot = true;
}

void Config::setAmiga5VPresence(bool state)
{
	amiga5VPresence = state;
}

void Config::setTmpAp()
{
	File file = LITTLEFS.open("/tmpAp", "w");
	if (!file)
	{
		Serial.println("Failed to save tmpAp file");
		return;
	}

	file.println("");
	file.close();
	tmpAp = true;
}

void Config::resetTmpAp()
{
	LITTLEFS.remove("/tmpAp");
}

bool Config::isTmpAp()
{
	File file = LITTLEFS.open("/tmpAp");
	if (!file || file.isDirectory())
	{
		file.close();
		return false;
	}
	return true;
}
