#include <Arduino.h>
#include <WiFi.h>
#include <ESPmDNS.h>

#include "main.h"
#include "networking.h"
#include "webPage.h"

IPAddress local_ip(192, 168, 4, 1);
IPAddress gateway(192, 168, 4, 1);
IPAddress subnet(255, 255, 255, 0);

// ugly method to get rid of callbacks without context
static Networking *currentThis = nullptr;

Networking::Networking()
{
	currentThis = this;
	wsClientConnected = false;
}

static void onIndexRequestStatic(AsyncWebServerRequest *request)
{
	currentThis->onIndexRequest(request);
}

static void handle_NotFoundStatic(AsyncWebServerRequest *request)
{
	currentThis->handle_notFound(request);
}

static void onWsEventStatic(AsyncWebSocket *server, AsyncWebSocketClient *client, AwsEventType type, void *arg, uint8_t *data, size_t len)
{
	currentThis->onWsEvent(server, client, type, arg, data, len);
}

void Networking::setup()
{
	WiFi.persistent(false);
	if (config->isAp)
	{
		Serial.println("Setting AP (Access Point)…");
		Serial.print("Access point SSID:");
		Serial.println(config->ssid);
		WiFi.softAP(config->ssid.c_str(), config->password.c_str());
		WiFi.softAPConfig(local_ip, gateway, subnet);
	}
	else if (config->tmpAp)
	{
		Serial.println("Setting AP (Access Point)…");
		String defSsid = defaultSsid;
		Serial.print("Access point SSID:");
		Serial.println(defSsid);
		WiFi.softAP(defSsid.c_str(), DEFPASSWD);
		WiFi.softAPConfig(local_ip, gateway, subnet);
		Serial.print("IP Address: ");
		Serial.println(WiFi.softAPIP());
		config->resetTmpAp(); // just this time
	}
	else
	{
		int timeoutCounter = 0;
		WiFi.begin(config->ssid.c_str(), config->password.c_str());
		while (WiFi.status() != WL_CONNECTED && ++timeoutCounter < WIFI_CON_TIMEOUT_COUNTER)
		{
			delay(480);
			digitalWrite(BLUELED, HIGH); // blue led on
			delay(20);
			digitalWrite(BLUELED, LOW); // blue led off
			Serial.println("Connecting to WiFi..");
		}
		if (timeoutCounter >= WIFI_CON_TIMEOUT_COUNTER)
		{
			// up a default AP
			// at next boot
			// WiFi.disconnect();
			Serial.print("Timeout connecting to ");
			Serial.println(config->ssid);
			Serial.println("Setting AP (Access Point) at next boot…");
			// tell to do AP next boot
			config->setTmpAp();
			// and reboot
			delay(1000);
			ESP.restart();
		}
		else
		{
			Serial.print("IP Address: ");
			Serial.println(WiFi.localIP());
		}
	}
	delay(100);
	Serial.println("Wifi started");
	if (!MDNS.begin(config->netName))
	{
		Serial.println("Cannot start mDNS");
	}

	server = new AsyncWebServer(80);
	server->on("/", HTTP_GET, onIndexRequestStatic);
	server->onNotFound(handle_NotFoundStatic);

	ws = new AsyncWebSocket("/ws");
	ws->onEvent(onWsEventStatic);
	server->addHandler(ws);

	server->begin();
	Serial.println("HTTP server started");
}

void Networking::setConfig(Config *conf)
{
	config = conf;
}

void Networking::setDefaultSsid(String val)
{
	defaultSsid = val;
}

void Networking::onIndexRequest(AsyncWebServerRequest *request)
{
	String str;
	Serial.println("Web asked");
	request->send(200, "text/html", index_html);
}

void Networking::handle_notFound(AsyncWebServerRequest *request)
{
	request->send(404, "text/plain", "Not found");
}

void Networking::onWsEvent(AsyncWebSocket *server, AsyncWebSocketClient *client, AwsEventType type, void *arg, uint8_t *data, size_t len)
{
	if (type == WS_EVT_CONNECT)
	{
		Serial.println("Websocket client connection received");
		wsClientConnected++;
		sendWsStates(client);
		return;
	}
	else if (type == WS_EVT_DISCONNECT)
	{
		Serial.println("Websocket client disconnected");
		wsClientConnected--;
		return;
	}
	else if (type == WS_EVT_DATA)
	{
		AwsFrameInfo *info = (AwsFrameInfo *)arg;
		String msg = "";
		if (info->final && info->index == 0 && info->len == len)
		{
			// Serial.printf("ws[%s][%u] %s-message[%llu]: ", server->url(), client->id(), (info->opcode == WS_TEXT)?"text":"binary", info->len);
			if (info->opcode != WS_TEXT)
			{
				Serial.println("Ws Binary mode not supported");
			}
			for (size_t i = 0; i < info->len; i++)
			{
				msg += (char)data[i];
			}
			Serial.printf("%s\n", msg.c_str());
			JsonDocument doc;
			DeserializationError error = deserializeJson(doc, msg);
			if (error)
			{
				Serial.print(F("deserializeJson() failed: "));
				Serial.println(error.c_str());
				return;
			}
			if (doc.containsKey("PING"))
			{
				// answer
				client->text("{\"PONG\":true}\n");
			}
			if (config->parseFromJson(doc))
			{
				ws->textAll(msg);
				config->saveConfig();
			}
		}
	}
}

void Networking::disconnectAll()
{
	ws->textAll("{\"EXIT\":true}\n");
	ws->closeAll();
}

void Networking::setAmiga5VPresence(bool state)
{
	if (lastAmiga5VPresent != state && ws && wsClientConnected)
	{
		if (state)
		{
			ws->textAll("{\"AMIGAPRES\":true}");
		}
		else
		{
			ws->textAll("{\"AMIGAPRES\":false}");
		}
	}
	lastAmiga5VPresent = state;
}

void Networking::sendWsStates(AsyncWebSocketClient *client)
{
	String jsonToSend;

	jsonToSend = config->toString();

	// Serial.println("WsSend:" + jsonToSend);

	if (client)
	{
		client->text(jsonToSend);
	}
}
